#include "cuda_check.cuh"
#include <math.h>

__global__ void matAdd(double *c, double *a, double *b, size_t n)
{
  size_t idx = blockIdx.x*blockDim.x+threadIdx.x;
  size_t idy = blockIdx.y*blockDim.y+threadIdx.y;
  size_t id;

  if (idx >= n || idy >= n) return;
  id = idx + idy * n;
  c[id] = a[id] + b[id];
}

__global__ void printU(double * a, size_t i, size_t j, size_t n)
{
  size_t k = i + j*n;
  printf("sur GPU %s[%ld,%ld] = %lg\n", "u", i, j, a[k]);
}

__global__ void printV(double * a, size_t i, size_t j, size_t n)
{
  size_t k = i + j*n;
  printf("sur GPU %s[%ld,%ld] = %lg\n", "v", i, j, a[k]);
}

int main(int argc, char **argv)
{
  size_t i, j, n = argc > 1 ? atoll(argv[1]) : 1000L;

  double *h_u, *h_v, *h_w;
  double *d_u, *d_v, *d_w;

/*
    Allocations memoire pour le CPU et le GPU
*/
  size_t bytes = n*n*sizeof(double);

  h_u = (double *) malloc(bytes);
  if (h_u == NULL) exit(-1);
  h_v = (double *) malloc(bytes);
  if (h_v == NULL) exit(-1);
  h_w = (double *) malloc(bytes);
  if (h_w == NULL) exit(-1);
  
  CUDA_CHECK_OP(cudaMalloc(&d_u, bytes));
  CUDA_CHECK_OP(cudaMalloc(&d_v, bytes));
  CUDA_CHECK_OP(cudaMalloc(&d_w, bytes));

/*
    Initialisation des donnees dans la memoire CPU
*/
  for (i=0; i<n; i++)
    for (j=0; j<n; j++)
    {
      size_t k = i + j*n;
      double x = i*1.0/n, y = j*1.0/n;

      h_u[k] = sin(x*y);
      h_v[k] = cos(x*y);
      h_w[k] = 0.0;
    }

/*
    Copie des donnees de la memoire CPU vers la memoire GPU
*/
  cudaMemcpy(d_u, h_u, bytes, cudaMemcpyHostToDevice);
  cudaMemcpy(d_v, h_v, bytes, cudaMemcpyHostToDevice);

/*
    Calcul sur GPU
*/
  int g = 16;
  int m = ceil(((double) n)/16);

  dim3 blockSize(g,g,1);
  dim3 gridSize(m,m,1);

  matAdd<<<gridSize, blockSize>>>(d_w, d_u, d_v, n);

  cudaDeviceSynchronize();
  CUDA_CHECK_KERNEL();

/*
    Copie des resultats de la memoire GPU vers la memoire CPU
*/
  cudaMemcpy(h_w, d_w, bytes, cudaMemcpyDeviceToHost);

/*
    Verification
*/
  printU<<<1,1>>>(d_u, n/2, n/3, n);
  printV<<<1,1>>>(d_v, n/2, n/3, n);
  cudaDeviceSynchronize();
  
  size_t k = n/2 + n*(n/3);

  printf("Sur GPU w[%ld,%ld] = %g\n", n/2, n/3, h_w[k]);
  printf("Sur CPU w[%ld,%ld] = %g\n", n/2, n/3, h_u[k] + h_v[k]);

/*
    Liberation de la memoire CPU et GPU
*/

  cudaFree(d_w);
  cudaFree(d_v);
  cudaFree(d_w);

  free(h_w);
  free(h_v);
  free(h_u);

  return 0;
}