#include "cuda_check.cuh"
#include <math.h>

struct matrice
{
  double **p;
  double *c;
  int n;
};

__global__ void matAdd(double **c, double **a, double **b, size_t n)
{
  size_t idx = blockIdx.x*blockDim.x+threadIdx.x;
  size_t idy = blockIdx.y*blockDim.y+threadIdx.y;

  if (idx >= n || idy >= n) return;
  c[idx][idy] = a[idx][idy] + b[idx][idy];
}

__global__ void printU(double ** a, size_t i, size_t j)
{
  printf("sur GPU %s[%ld,%ld] = %lg\n", "u", i, j, a[i][j]);
}

__global__ void printV(double ** a, size_t i, size_t j)
{
  printf("sur GPU %s[%ld,%ld] = %lg\n", "v", i, j, a[i][j]);
}

void alloueCPU(matrice * m, int n)
{
  size_t pbytes = n*sizeof(double*);
  size_t bytes = n*n*sizeof(double);
  size_t j;

  m->n = n;
  m->p = (double **) malloc(pbytes);
  m->c = (double *) malloc(bytes); 

  for (j=0; j<n; j++)
    m->p[j] = m->c + j*n;
}

void libereCPU(matrice *m)
{
  free(m->p);
  free(m->c);
}

__global__ void connecte(double **p, double *c, int n)
{
  size_t id = blockIdx.x*blockDim.x+threadIdx.x;
  if (id < n)
    p[id] = c + id*n;
}

void alloueGPU(matrice *m, int n)
{
  size_t pbytes = n*sizeof(double*);
  size_t bytes = n*n*sizeof(double);
  
  m->n = n;
  CUDA_CHECK_OP(cudaMalloc(&(m->c), bytes));
  CUDA_CHECK_OP(cudaMalloc(&(m->p), pbytes));

  connecte<<<256, ceil((double) n/256)>>>(m->p, m->c, m->n);
  cudaDeviceSynchronize();
  CUDA_CHECK_KERNEL();
}

void libereGPU(matrice * m)
{
  cudaFree(m->p);
  cudaFree(m->c);
}

int main(int argc, char **argv)
{
  size_t i, j, n = argc > 1 ? atoll(argv[1]) : 1000L;

  matrice h_mu, h_mv, h_mw;
  matrice d_mu, d_mv, d_mw;
  
/*
    Allocations memoire pour le CPU et le GPU
*/

  alloueCPU(&h_mu, n);
  alloueCPU(&h_mv, n);
  alloueCPU(&h_mw, n);
    
  alloueGPU(&d_mu, n);
  alloueGPU(&d_mv, n);
  alloueGPU(&d_mw, n);

/*
    Initialisation des donnees dans la memoire CPU
*/
  for (i=0; i<n; i++)
    for (j=0; j<n; j++)
    {
      double x = i*1.0/n, y = j*1.0/n;

      h_mu.p[i][j] = sin(x*y);
      h_mv.p[i][j] = cos(x*y);
      h_mw.p[i][j] = 0.0;
    }

/*
    Copie des donnees de la memoire CPU vers la memoire GPU
*/
  size_t bytes = n*n*sizeof(double);
  cudaMemcpy(d_mu.c, h_mu.c, bytes, cudaMemcpyHostToDevice);
  cudaMemcpy(d_mv.c, h_mv.c, bytes, cudaMemcpyHostToDevice);
/*
    Calcul sur GPU
*/
  int g = 16;
  int m = ceil(((double) n)/16);

  dim3 blockSize(g,g,1);
  dim3 gridSize(m,m,1);

  matAdd<<<gridSize, blockSize>>>(d_mw.p, d_mu.p, d_mv.p, n);

  cudaDeviceSynchronize();
  CUDA_CHECK_KERNEL();

/*
    Copie des resultats de la memoire GPU vers la memoire CPU
*/
  cudaMemcpy(h_mw.c, d_mw.c, bytes, cudaMemcpyDeviceToHost);

/*
    Verification
*/
  printU<<<1,1>>>(d_mu.p, n/2, n/3);
  printV<<<1,1>>>(d_mv.p, n/2, n/3);
  cudaDeviceSynchronize();
  
  printf("Sur GPU w[%ld,%ld] = %g\n", n/2, n/3,
         h_mw.p[n/2][n/3]);
  printf("Sur CPU w[%ld,%ld] = %g\n", n/2, n/3, 
         h_mu.p[n/2][n/3] + h_mv.p[n/2][n/3]);

/*
    Liberation de la memoire CPU et GPU
*/
  libereGPU(&d_mw);
  libereGPU(&d_mv);
  libereGPU(&d_mu);

  libereCPU(&h_mw);
  libereCPU(&h_mv);
  libereCPU(&h_mu);

  return 0;
}