#include "cuda_check.cuh"
#include <math.h>

__global__ void vecAdd(double *c, double *a, double *b, size_t n)
{
  size_t id = blockIdx.x*blockDim.x+threadIdx.x;
  
  if (id >= n) return;
  c[id] = a[id] + b[id];
}

__global__ void printU(double * a, size_t i)
{
    printf("sur GPU %s[%ld] = %lg\n", "u", i, a[i]);
}

__global__ void printV(double * a, size_t i)
{
    printf("sur GPU %s[%ld] = %lg\n", "v", i, a[i]);
}

int main(int argc, char **argv)
{
  unsigned long i, n = argc > 1 ? atoll(argv[1]) : 1000000L;

  double *h_u, *h_v, *h_w;
  double *d_u, *d_v, *d_w;

/*
    Allocations memoire pour le CPU et le GPU
*/
  size_t bytes = n*sizeof(double);

  h_u = (double *) malloc(bytes);
  if (h_u == NULL) exit(-1);
  h_v = (double *) malloc(bytes);
  if (h_v == NULL) exit(-1);
  h_w = (double *) malloc(bytes);
  if (h_w == NULL) exit(-1);
  
  CUDA_CHECK_OP(cudaMalloc(&d_u, bytes));
  CUDA_CHECK_OP(cudaMalloc(&d_v, bytes));
  CUDA_CHECK_OP(cudaMalloc(&d_w, bytes));

/*
    Initialisation des donnees dans la memoire CPU
*/
  for (i=0; i<n; i++)
  {
    h_u[i] = sin(i*1.0/n);
    h_v[i] = cos(i*1.0/n);
    h_w[i] = 0.0;
  }

/*
    Copie des donnees de la memoire CPU vers la memoire GPU
*/
  cudaMemcpy(d_u, h_u, bytes, cudaMemcpyHostToDevice);
  cudaMemcpy(d_v, h_v, bytes, cudaMemcpyHostToDevice);

/*
    Calcul sur GPU
*/
  unsigned int blockSize = 512;
  unsigned int gridSize = (unsigned int) ceil((double)n/blockSize);

  vecAdd<<<gridSize, blockSize>>>(d_w, d_u, d_v, n);

/*
    Copie des resultats de la memoire GPU vers la memoire CPU
*/
  cudaMemcpy(h_w, d_w, bytes, cudaMemcpyDeviceToHost);

/*
    Verification
*/
  printU<<<1,1>>>(d_u, n/2);
  printV<<<1,1>>>(d_v, n/2);
  cudaDeviceSynchronize();
  CUDA_CHECK_KERNEL();

  printf("Sur GPU w[%ld] = %g\n", n/2, h_w[n/2]);
  printf("Sur CPU w[%ld] = %g\n", n/2, h_u[n/2] + h_v[n/2]);

/*
    Liberation de la memoire CPU et GPU
*/

  cudaFree(d_w);
  cudaFree(d_v);
  cudaFree(d_w);

  free(h_w);
  free(h_v);
  free(h_u);

  return 0;
}